import base64
from datetime import date, datetime
from enum import Enum

from pygrocy2.base import DataModel
from pygrocy2.grocy_api_client import (
    GrocyApiClient,
    MealPlanResponse,
    MealPlanSectionResponse,
    RecipeDetailsResponse,
)


class RecipeItem(DataModel):
    def __init__(self, response: RecipeDetailsResponse):
        self._id = response.id
        self._name = response.name
        self._description = response.description
        self._base_servings = response.base_servings
        self._desired_servings = response.desired_servings
        self._picture_file_name = response.picture_file_name

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description

    @property
    def base_servings(self) -> int:
        return self._base_servings

    @property
    def desired_servings(self) -> int:
        return self._desired_servings

    @property
    def picture_file_name(self) -> str:
        return self._picture_file_name

    def get_picture_url_path(self, width: int = 400):
        if self.picture_file_name:
            b64name = base64.b64encode(self.picture_file_name.encode("ascii"))
            path = "files/recipepictures/" + str(b64name, "utf-8")

            return f"{path}?force_serve_as=picture&best_fit_width={width}"


class MealPlanSection(DataModel):
    def __init__(self, response: MealPlanSectionResponse):
        self._id = response.id
        self._name = response.name
        self._sort_number = response.sort_number
        self._row_created_timestamp = response.row_created_timestamp

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def sort_number(self) -> int:
        return self._sort_number

    @property
    def row_created_timestamp(self) -> datetime:
        return self._row_created_timestamp


class MealPlanItemType(str, Enum):
    NOTE = "note"
    PRODUCT = "product"
    RECIPE = "recipe"


class MealPlanItem(DataModel):
    def __init__(self, response: MealPlanResponse):
        self._id = response.id
        self._day = self._normalize_day(response.day)
        self._recipe = None
        self._recipe_id = response.recipe_id
        self._recipe_servings = response.recipe_servings
        self._note = response.note
        self._section_id = response.section_id
        self._type = MealPlanItemType(response.type)
        self._product_id = response.product_id

    @property
    def id(self) -> int:
        return self._id

    @property
    def day(self) -> date | None:
        return self._day

    @property
    def recipe_id(self) -> int:
        return self._recipe_id

    @property
    def recipe_servings(self) -> int:
        return self._recipe_servings

    @property
    def note(self) -> str:
        return self._note

    @property
    def recipe(self) -> RecipeItem:
        return self._recipe

    @property
    def section_id(self) -> int:
        return self._section_id

    @property
    def section(self) -> MealPlanSection:
        return self._section

    @property
    def type(self) -> MealPlanItemType:
        return self._type

    @property
    def product_id(self) -> int:
        return self._product_id

    def get_details(self, api_client: GrocyApiClient):
        if self.recipe_id:
            recipe = api_client.get_recipe(self.recipe_id)
            if recipe:
                self._recipe = RecipeItem(recipe)
        if self.section_id:
            section = api_client.get_meal_plan_section(self.section_id)
            if section:
                self._section = MealPlanSection(section)

    @staticmethod
    def _normalize_day(value) -> date | None:
        if value is None:
            return None
        if isinstance(value, date) and not isinstance(value, datetime):
            return value
        if isinstance(value, datetime):
            return value.date()
        # Grocy sometimes returns ISO8601 strings in this field.
        return datetime.fromisoformat(str(value)).date()



from pygrocy2.grocy_api_client import GrocyApiClient


class TestGrocyApiClient:
    def test_url_only(self):
        client = GrocyApiClient(api_key="", base_url="http://grocy.de")
        assert client._base_url == "http://grocy.de:9192/api/"

    def test_url_and_port(self):
        client = GrocyApiClient(api_key="", base_url="http://grocy.de", port=1234)
        assert client._base_url == "http://grocy.de:1234/api/"

    def test_url_and_port_and_path(self):
        client = GrocyApiClient(
            api_key="", base_url="http://grocy.de", port=1234, path="my/custom/path"
        )
        assert client._base_url == "http://grocy.de:1234/my/custom/path/api/"

    def test_url_and_path(self):
        client = GrocyApiClient(
            api_key="", base_url="http://grocy.de", path="my/custom/path"
        )
        assert client._base_url == "http://grocy.de:9192/my/custom/path/api/"



import logging
from datetime import datetime

import deprecation

from .base import DataModel  # noqa: F401
from .data_models.battery import Battery
from .data_models.chore import Chore
from .data_models.equipment import Equipment
from .data_models.generic import EntityType
from .data_models.meal_items import MealPlanItem, MealPlanSection, RecipeItem
from .data_models.product import Group, Product, ShoppingListProduct
from .data_models.system import SystemConfig, SystemInfo, SystemTime
from .data_models.task import Task
from .data_models.user import User  # noqa: F401
from .errors import GrocyError  # noqa: F401
from .grocy_api_client import ChoreDetailsResponse  # noqa: F401
from .grocy_api_client import CurrentChoreResponse  # noqa: F401
from .grocy_api_client import CurrentStockResponse  # noqa: F401
from .grocy_api_client import LocationData  # noqa: F401
from .grocy_api_client import MealPlanResponse  # noqa: F401
from .grocy_api_client import MissingProductResponse  # noqa: F401
from .grocy_api_client import ProductDetailsResponse  # noqa: F401
from .grocy_api_client import RecipeDetailsResponse  # noqa: F401
from .grocy_api_client import ShoppingListItem  # noqa: F401
from .grocy_api_client import TaskResponse  # noqa: F401
from .grocy_api_client import UserDto  # noqa: F401
from .grocy_api_client import DEFAULT_PORT_NUMBER, GrocyApiClient, TransactionType

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


class Grocy(object):
    def __init__(
        self,
        base_url,
        api_key,
        port: int = DEFAULT_PORT_NUMBER,
        path: str | None = None,
        verify_ssl=True,
        debug=False,
    ):
        self._api_client = GrocyApiClient(
            base_url, api_key, port, path, verify_ssl, debug
        )

        if debug:
            _LOGGER.setLevel(logging.DEBUG)

    def stock(self) -> list[Product]:
        raw_stock = self._api_client.get_stock()
        return [Product(resp) for resp in raw_stock]

    @deprecation.deprecated(details="Use due_products instead")
    def expiring_products(self, get_details: bool = False) -> list[Product]:
        return self.due_products(get_details)

    def due_products(self, get_details: bool = False) -> list[Product]:
        raw_due_products = self._api_client.get_volatile_stock().due_products
        due_products = [Product(resp) for resp in raw_due_products]

        if get_details:
            for item in due_products:
                item.get_details(self._api_client)
        return due_products

    def overdue_products(self, get_details: bool = False) -> list[Product]:
        raw_overdue_products = self._api_client.get_volatile_stock().overdue_products
        overdue_products = [Product(resp) for resp in raw_overdue_products]

        if get_details:
            for item in overdue_products:
                item.get_details(self._api_client)
        return overdue_products

    def expired_products(self, get_details: bool = False) -> list[Product]:
        raw_expired_products = self._api_client.get_volatile_stock().expired_products
        expired_products = [Product(resp) for resp in raw_expired_products]

        if get_details:
            for item in expired_products:
                item.get_details(self._api_client)
        return expired_products

    def missing_products(self, get_details: bool = False) -> list[Product]:
        raw_missing_products = self._api_client.get_volatile_stock().missing_products
        missing_products = [Product(resp) for resp in raw_missing_products]

        if get_details:
            for item in missing_products:
                item.get_details(self._api_client)
        return missing_products

    def product(self, product_id: int) -> Product:
        resp = self._api_client.get_product(product_id)
        if resp:
            return Product(resp)
        return None

    def product_by_barcode(self, barcode: str) -> Product:
        resp = self._api_client.get_product_by_barcode(barcode)
        if resp:
            return Product(resp)
        return None

    def all_products(self) -> list[Product]:
        raw_products = self.get_generic_objects_for_type(EntityType.PRODUCTS)
        from pygrocy2.grocy_api_client import ProductData

        product_datas = [ProductData(**product) for product in raw_products]
        return [Product(product) for product in product_datas]

    def chores(
        self, get_details: bool = False, query_filters: list[str] | None = None
    ) -> list[Chore]:
        raw_chores = self._api_client.get_chores(query_filters)
        chores = [Chore(chore) for chore in raw_chores]

        if get_details:
            for chore in chores:
                chore.get_details(self._api_client)
        return chores

    def execute_chore(
        self,
        chore_id: int,
        done_by: int = None,
        tracked_time: datetime = None,
        skipped: bool = False,
    ):
        return self._api_client.execute_chore(chore_id, done_by, tracked_time, skipped)

    def chore(self, chore_id: int) -> Chore:
        resp = self._api_client.get_chore(chore_id)
        return Chore(resp)

    def add_product(
        self,
        product_id,
        amount: float,
        price: float,
        best_before_date: datetime | None = None,
        transaction_type: TransactionType = TransactionType.PURCHASE,
    ):
        return self._api_client.add_product(
            product_id, amount, price, best_before_date, transaction_type
        )

    def consume_product(
        self,
        product_id: int,
        amount: float = 1,
        spoiled: bool = False,
        transaction_type: TransactionType = TransactionType.CONSUME,
        allow_subproduct_substitution: bool = False,
    ):
        return self._api_client.consume_product(
            product_id, amount, spoiled, transaction_type, allow_subproduct_substitution
        )

    def consume_recipe(
        self,
        recipe_id: int,
    ):
        return self._api_client.consume_recipe(recipe_id)

    def open_product(
        self,
        product_id: int,
        amount: float = 1,
        allow_subproduct_substitution: bool = False,
    ):
        return self._api_client.open_product(
            product_id, amount, allow_subproduct_substitution
        )

    def inventory_product(
        self,
        product_id: int,
        new_amount: float,
        best_before_date: datetime | None = None,
        shopping_location_id: int | None = None,
        location_id: int | None = None,
        price: float | None = None,
        get_details: bool = True,
    ) -> Product:
        product = Product(
            self._api_client.inventory_product(
                product_id,
                new_amount,
                best_before_date,
                shopping_location_id,
                location_id,
                price,
            )
        )

        if get_details:
            product.get_details(self._api_client)
        return product

    def add_product_by_barcode(
        self,
        barcode: str,
        amount: float,
        price: float,
        best_before_date: datetime | None = None,
        get_details: bool = True,
    ) -> Product:
        product = Product(
            self._api_client.add_product_by_barcode(
                barcode, amount, price, best_before_date
            )
        )

        if get_details:
            product.get_details(self._api_client)
        return product

    def consume_product_by_barcode(
        self,
        barcode: str,
        amount: float = 1,
        spoiled: bool = False,
        get_details: bool = True,
    ) -> Product:
        product = Product(
            self._api_client.consume_product_by_barcode(barcode, amount, spoiled)
        )

        if get_details:
            product.get_details(self._api_client)
        return product

    def inventory_product_by_barcode(
        self,
        barcode: str,
        new_amount: float,
        best_before_date: datetime | None = None,
        location_id: int | None = None,
        price: float | None = None,
        get_details: bool = True,
    ) -> Product:
        product = Product(
            self._api_client.inventory_product_by_barcode(
                barcode, new_amount, best_before_date, location_id, price
            )
        )

        if get_details:
            product.get_details(self._api_client)
        return product

    def shopping_list(
        self, get_details: bool = False, query_filters: list[str] | None = None
    ) -> list[ShoppingListProduct]:
        raw_shoppinglist = self._api_client.get_shopping_list(query_filters)
        shopping_list = [ShoppingListProduct(resp) for resp in raw_shoppinglist]

        if get_details:
            for item in shopping_list:
                item.get_details(self._api_client)
        return shopping_list

    def add_missing_product_to_shopping_list(self, shopping_list_id: int = 1):
        return self._api_client.add_missing_product_to_shopping_list(shopping_list_id)

    def add_product_to_shopping_list(
        self,
        product_id: int,
        shopping_list_id: int | None = None,
        amount: float | None = None,
        quantity_unit_id: int | None = None,
    ):
        return self._api_client.add_product_to_shopping_list(
            product_id, shopping_list_id, amount, quantity_unit_id
        )

    def clear_shopping_list(self, shopping_list_id: int = 1):
        return self._api_client.clear_shopping_list(shopping_list_id)

    def remove_product_in_shopping_list(
        self, product_id: int, shopping_list_id: int = 1, amount: float = 1
    ):
        return self._api_client.remove_product_in_shopping_list(
            product_id, shopping_list_id, amount
        )

    def product_groups(self, query_filters: list[str] | None = None) -> list[Group]:
        raw_groups = self._api_client.get_product_groups(query_filters)
        return [Group(resp) for resp in raw_groups]

    def add_product_pic(self, product_id: int, pic_path: str):
        self._api_client.upload_product_picture(product_id, pic_path)
        return self._api_client.update_product_pic(product_id)

    def get_userfields(self, entity: str, object_id: int):
        return self._api_client.get_userfields(entity, object_id)

    def set_userfields(self, entity: str, object_id: int, key: str, value):
        return self._api_client.set_userfields(entity, object_id, key, value)

    def get_last_db_changed(self):
        return self._api_client.get_last_db_changed()

    def get_system_info(self) -> SystemInfo:
        raw_system_info = self._api_client.get_system_info()
        if raw_system_info:
            return SystemInfo(raw_system_info)
        return None

    def get_system_time(self) -> SystemTime:
        raw_system_time = self._api_client.get_system_time()
        if raw_system_time:
            return SystemTime(raw_system_time)
        return None

    def get_system_config(self) -> SystemConfig:
        raw_system_config = self._api_client.get_system_config()
        if raw_system_config:
            return SystemConfig(raw_system_config)
        return None

    def tasks(self, query_filters: list[str] | None = None) -> list[Task]:
        raw_tasks = self._api_client.get_tasks(query_filters)
        return [Task(task) for task in raw_tasks]

    def task(self, task_id: int) -> Task:
        resp = self._api_client.get_task(task_id)
        return Task(resp)

    def complete_task(self, task_id, done_time: datetime | None = None):
        return self._api_client.complete_task(task_id, done_time)

    def meal_plan(
        self, get_details: bool = False, query_filters: list[str] | None = None
    ) -> list[MealPlanItem]:
        raw_meal_plan = self._api_client.get_meal_plan(query_filters)
        meal_plan = [MealPlanItem(data) for data in raw_meal_plan]

        if get_details:
            for item in meal_plan:
                item.get_details(self._api_client)
        return meal_plan

    def recipe(self, recipe_id: int) -> RecipeItem:
        recipe = self._api_client.get_recipe(recipe_id)
        if recipe:
            return RecipeItem(recipe)
        return None

    def batteries(
        self, query_filters: list[str] | None = None, get_details: bool = False
    ) -> list[Battery]:
        raw_batteries = self._api_client.get_batteries(query_filters)
        batteries = [Battery(bat) for bat in raw_batteries]

        if get_details:
            for item in batteries:
                item.get_details(self._api_client)
        return batteries

    def battery(self, battery_id: int) -> Battery:
        battery = self._api_client.get_battery(battery_id)
        if battery:
            return Battery(battery)
        return None

    def charge_battery(self, battery_id: int, tracked_time: datetime | None = None):
        return self._api_client.charge_battery(battery_id, tracked_time)

    def add_generic(self, entity_type: EntityType, data):
        return self._api_client.add_generic(entity_type.value, data)

    def get_generic(self, entity_type: EntityType, object_id: int):
        return self._api_client.get_generic(entity_type.value, object_id)

    def update_generic(self, entity_type: EntityType, object_id: int, updated_data):
        return self._api_client.update_generic(
            entity_type.value, object_id, updated_data
        )

    def delete_generic(self, entity_type: EntityType, object_id: int):
        return self._api_client.delete_generic(entity_type.value, object_id)

    def get_generic_objects_for_type(
        self, entity_type: EntityType, query_filters: list[str] | None = None
    ):
        return self._api_client.get_generic_objects_for_type(
            entity_type.value, query_filters
        )

    def meal_plan_sections(
        self, query_filters: list[str] | None = None
    ) -> list[MealPlanSection]:
        raw_sections = self._api_client.get_meal_plan_sections(query_filters)
        return [MealPlanSection(section) for section in raw_sections]

    def meal_plan_section(self, meal_plan_section_id: int) -> MealPlanSection:
        section = self._api_client.get_meal_plan_section(meal_plan_section_id)

        if section:
            return MealPlanSection(section)
        return None

    def users(self) -> list[User]:
        user_dtos = self._api_client.get_users()
        return [User(user) for user in user_dtos]

    def user(self, user_id: int | None = None) -> User:
        user = self._api_client.get_user(user_id=user_id)
        if user:
            return User(user)
        return None

    def equipment(
        self, query_filters: list[str] = None, get_details: bool = False
    ) -> list[Equipment]:
        """Return a list of equipment items."""
        raw_equipment = self._api_client.get_all_equipment(query_filters)
        equipment_items = [Equipment(item) for item in raw_equipment]

        if get_details:
            for item in equipment_items:
                item.get_details(self._api_client)

        return equipment_items

    def get_equipment(self, equipment_id: int) -> Equipment:
        """Return a single equipment item by ID."""
        equipment_data = self._api_client.get_equipment(equipment_id)
        if equipment_data:
            return Equipment(equipment_data)
        return None

    def get_equipment_by_name(self, name: str) -> Equipment:
        """Return a single equipment item by name."""
        query_filters = [f"name={name}"]
        equipment_items = self.equipment(query_filters, True)
        if equipment_items and len(equipment_items) > 0:
            return equipment_items[0]
        return None

    def get_all_equipment_objects(self) -> list[Equipment]:
        """Return all equipment objects with full details."""
        raw_equipment = self.get_generic_objects_for_type(EntityType.EQUIPMENT)

        from .grocy_api_client import EquipmentData

        equipment_data = [EquipmentData(**equipment) for equipment in raw_equipment]

        equipment_items = []
        for item in equipment_data:
            equipment_details = self._api_client.get_equipment(item.id)
            if equipment_details:
                equipment_items.append(Equipment(equipment_details))

        return equipment_items



from datetime import datetime

import zoneinfo

import pygrocy2.utils as utils


class TestUtils:
    def test_parse_date_valid(self):
        date_str = "2022-07-10 21:10:53"
        date_obj = utils.parse_date(date_str)

        assert isinstance(date_obj, datetime)

    def test_parse_date_no_data(self):
        date_str = None
        date_obj = utils.parse_date(date_str)

        assert date_obj is None

    def test_parse_date_empty_string(self):
        date_str = ""
        date_obj = utils.parse_date(date_str)

        assert date_obj is None

    def test_parse_int_valid(self):
        int_str = "2"
        int_number = utils.parse_int(int_str)

        assert isinstance(int_number, int)

    def test_parse_int_no_data(self):
        int_str = None
        int_number = utils.parse_int(int_str, -1)

        assert int_number == -1

    def test_parse_int_error(self):
        int_str = "string"
        int_number = utils.parse_int(int_str, -1)

        assert int_number == -1

    def test_parse_float_valid(self):
        float_str = "2.01"
        float_number = utils.parse_float(float_str)

        assert isinstance(float_number, float)

    def test_parse_float_no_data(self):
        float_str = None
        float_number = utils.parse_float(float_str, -1)

        assert float_number == -1

    def test_parse_float_error(self):
        float_str = "string"
        float_number = utils.parse_float(float_str, -1)

        assert float_number == -1

    def test_localize_datetime_input_timezone_unaware(self):
        date = datetime(2022, 7, 10, 21, 17, 34, 633809, tzinfo=None)

        localized_datetime = utils.localize_datetime(date)

        assert localized_datetime.replace(tzinfo=None) == date
        assert localized_datetime.tzinfo is not None

    def test_localize_datetime_input_timezone_aware(self):
        los_angeles = zoneinfo.ZoneInfo("America/Los_Angeles")

        date = datetime(
            2022,
            7,
            10,
            13,
            17,
            34,
            633809,
            tzinfo=los_angeles,
        )

        localized_datetime = utils.localize_datetime(date)

        assert localized_datetime == datetime(
            2022,
            7,
            10,
            13,
            17,
            34,
            633809,
            tzinfo=los_angeles,
        )

    def test_grocy_datetime_str(self):
        date = datetime(2022, 7, 10, 21, 17, 34, 633809, tzinfo=None)
        date_str = utils.grocy_datetime_str(date)

        assert date_str == "2022-07-10 21:17:34"



import base64  # noqa: D100
from datetime import datetime
from enum import Enum
import json
import logging
from typing import Any
from urllib.parse import urljoin

from pydantic import BaseModel, Field, field_validator, model_validator
import requests

from .data_models.generic import EntityType
from .errors import GrocyError
from .utils import grocy_datetime_str, localize_datetime, parse_date

DEFAULT_PORT_NUMBER = 9192

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


def _field_not_empty_validator(field_name: str):
    """Reusable Pydantic field pre-validator to convert empty str to None."""
    return field_validator(field_name, mode="before")(_none_if_empty_str)


def _none_if_empty_str(value: Any):
    if isinstance(value, str) and value == "":
        return None
    return value


class ShoppingListItem(BaseModel):
    id: int
    product_id: int | None = None
    note: str | None = None
    amount: float | None = None
    row_created_timestamp: datetime
    shopping_list_id: int
    done: int


class MealPlanResponse(BaseModel):
    id: int
    day: datetime
    type: str
    recipe_id: int | None = None
    recipe_servings: int | None = None
    note: str | None = None
    product_id: int | None = None
    product_amount: float | None = None
    product_qu_id: str | None = None
    row_created_timestamp: datetime
    userfields: dict | None = None
    section_id: int | None = None


class RecipeDetailsResponse(BaseModel):
    id: int | None = None
    name: str
    description: str | None = None
    base_servings: int
    desired_servings: int
    picture_file_name: str | None = None
    row_created_timestamp: datetime
    userfields: dict | None = None


class QuantityUnitData(BaseModel):
    id: int
    name: str
    name_plural: str | None = None
    description: str | None = None
    row_created_timestamp: datetime


class LocationData(BaseModel):
    id: int
    name: str
    description: str | None = None
    row_created_timestamp: datetime


class ProductData(BaseModel):
    id: int
    name: str
    description: str | None = None
    location_id: int | None = None
    product_group_id: int | None = None
    qu_id_stock: int
    qu_id_purchase: int
    picture_file_name: str | None = None
    allow_partial_units_in_stock: bool | None = False
    row_created_timestamp: datetime
    min_stock_amount: float | None = None
    default_best_before_days: int

    location_id_validator = _field_not_empty_validator("location_id")
    product_group_id_validator = _field_not_empty_validator("product_group_id")


class ChoreData(BaseModel):
    id: int
    name: str
    description: str | None = None
    period_type: str
    period_config: str | None = None
    period_days: int | None = 0
    track_date_only: bool
    rollover: bool
    assignment_type: str | None = None
    assignment_config: str | None = None
    next_execution_assigned_to_user_id: int | None = None
    userfields: dict | None = None

    next_execution_assigned_to_user_id_validator = _field_not_empty_validator(
        "next_execution_assigned_to_user_id"
    )


class UserDto(BaseModel):
    id: int
    username: str
    first_name: str | None = None
    last_name: str | None = None
    display_name: str | None = None


class CurrentChoreResponse(BaseModel):
    chore_id: int
    last_tracked_time: datetime | None = None
    next_estimated_execution_time: datetime | None = None


class CurrentStockResponse(BaseModel):
    product_id: int
    amount: float
    best_before_date: datetime
    amount_opened: float
    amount_aggregated: float
    amount_opened_aggregated: float
    is_aggregated_amount: bool
    product: ProductData


class MissingProductResponse(BaseModel):
    id: int
    name: str
    amount_missing: float
    is_partly_in_stock: bool


class CurrentVolatilStockResponse(BaseModel):
    due_products: list[CurrentStockResponse] | None = None
    overdue_products: list[CurrentStockResponse] | None = None
    expired_products: list[CurrentStockResponse] | None = None
    missing_products: list[MissingProductResponse] | None = None


class ProductBarcodeData(BaseModel):
    barcode: str
    amount: float | None = None


class ProductDetailsResponse(BaseModel):
    last_purchased: datetime | None = None
    last_used: datetime | None = None
    stock_amount: float
    stock_amount_opened: float
    next_best_before_date: datetime | None = None
    last_price: float | None = None
    product: ProductData
    quantity_unit_stock: QuantityUnitData
    default_quantity_unit_purchase: QuantityUnitData
    barcodes: list[ProductBarcodeData] | None = Field(None, alias="product_barcodes")
    location: LocationData | None = None


class ChoreDetailsResponse(BaseModel):
    chore: ChoreData
    last_tracked: datetime | None = None
    next_estimated_execution_time: datetime | None = None
    track_count: int = 0
    next_execution_assigned_user: UserDto | None = None
    last_done_by: UserDto | None = None


class TransactionType(Enum):
    PURCHASE = "purchase"
    CONSUME = "consume"
    INVENTORY_CORRECTION = "inventory-correction"
    PRODUCT_OPENED = "product-opened"


class TaskCategoryDto(BaseModel):
    id: int
    name: str
    description: str | None = None
    row_created_timestamp: datetime


class TaskResponse(BaseModel):
    id: int
    name: str
    description: str | None = None
    due_date: datetime | None = None
    done: int
    done_timestamp: datetime | None = None
    category_id: int | None = None
    category: TaskCategoryDto | None = None
    assigned_to_user_id: int | None = None
    assigned_to_user: UserDto | None = None
    userfields: dict | None = None

    due_date_validator = _field_not_empty_validator("due_date")
    category_id_validator = _field_not_empty_validator("category_id")
    assigned_to_user_id_validator = _field_not_empty_validator("assigned_to_user_id")


class CurrentBatteryResponse(BaseModel):
    id: int
    last_tracked_time: datetime | None = None
    next_estimated_charge_time: datetime | None = None


class BatteryData(BaseModel):
    id: int
    name: str
    description: str | None = None
    used_in: str | None = None
    charge_interval_days: int
    created_timestamp: datetime = Field(alias="row_created_timestamp")
    userfields: dict | None = None


class BatteryDetailsResponse(BaseModel):
    battery: BatteryData
    charge_cycles_count: int
    last_charged: datetime | None = None
    last_tracked_time: datetime | None = None
    next_estimated_charge_time: datetime | None = None


class EquipmentData(BaseModel):
    id: int
    name: str
    description: str | None = None
    instruction_manual_file_name: str | None = None
    created_timestamp: datetime = Field(alias="row_created_timestamp")
    userfields: dict | None = None


class EquipmentDetailsResponse(BaseModel):
    equipment: EquipmentData


class MealPlanSectionResponse(BaseModel):
    id: int | None = None
    name: str | None = None
    sort_number: int | None = None
    row_created_timestamp: datetime

    sort_number_validator = _field_not_empty_validator("sort_number")


class StockLogResponse(BaseModel):
    id: int
    product_id: int
    amount: float
    best_before_date: datetime
    purchased_date: datetime
    used_date: datetime | None = None
    spoiled: bool = False
    stock_id: str
    transaction_id: str
    transaction_type: TransactionType


class GrocyVersionDto(BaseModel):
    version: str = Field(alias="Version")
    release_date: datetime = Field(alias="ReleaseDate")


class SystemInfoDto(BaseModel):
    grocy_version_info: GrocyVersionDto = Field(alias="grocy_version")
    php_version: str
    sqlite_version: str
    os: str
    client: str


class SystemTimeDto(BaseModel):
    timezone: str
    time_local: datetime
    time_local_sqlite3: datetime
    time_utc: datetime
    timestamp: int


class SystemConfigDto(BaseModel, extra="allow"):
    username: str = Field(alias="USER_USERNAME")
    base_path: str = Field(alias="BASE_PATH")
    base_url: str = Field(alias="BASE_URL")
    mode: str = Field(alias="MODE")
    default_locale: str = Field(alias="DEFAULT_LOCALE")
    locale: str = Field(alias="LOCALE")
    currency: str = Field(alias="CURRENCY")
    feature_flags: dict[str, Any]

    @model_validator(mode="before")
    @classmethod
    def feature_flags_root_validator(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Pydantic root validator to add all "FEATURE_FLAG_" settings to Dict."""
        features: dict[str, Any] = {}

        for field, value in data.items():
            if field.startswith("FEATURE_FLAG_"):
                features[field] = value

        data["feature_flags"] = features

        return data


def _enable_debug_mode():
    _LOGGER.setLevel(logging.DEBUG)


class GrocyApiClient(object):
    def __init__(
        self,
        base_url,
        api_key,
        port: int = DEFAULT_PORT_NUMBER,
        path: str | None = None,
        verify_ssl=True,
        debug=False,
    ):
        if debug:
            _enable_debug_mode()

        if path:
            self._base_url = f"{base_url}:{port}/{path}/api/"
        else:
            self._base_url = f"{base_url}:{port}/api/"
        _LOGGER.debug(f"generated base url: {self._base_url}")

        self._api_key = api_key
        self._verify_ssl = verify_ssl
        if self._api_key == "demo_mode":
            self._headers = {"accept": "application/json"}
        else:
            self._headers = {"accept": "application/json", "GROCY-API-KEY": api_key}

    def _do_get_request(self, end_url: str, query_filters: list[str] | None = None):
        req_url = urljoin(self._base_url, end_url)
        params = None
        if query_filters:
            params = {"query[]": query_filters}
        resp = requests.get(
            req_url, verify=self._verify_ssl, headers=self._headers, params=params
        )

        _LOGGER.debug("-->\tGET /%s", end_url)
        _LOGGER.debug("<--\t%d for /%s", resp.status_code, end_url)
        _LOGGER.debug("\t\t%s", resp.content)

        if resp.status_code >= 400:
            raise GrocyError(resp)

        if len(resp.content) > 0:
            return resp.json()
        return None

    def _do_post_request(self, end_url: str, data: dict):
        req_url = urljoin(self._base_url, end_url)
        resp = requests.post(
            req_url, verify=self._verify_ssl, headers=self._headers, json=data
        )

        _LOGGER.debug("-->\tPOST /%s", end_url)
        _LOGGER.debug("\t\t%s", data)
        _LOGGER.debug("<--\t%d for /%s", resp.status_code, end_url)
        _LOGGER.debug("\t\t%s", resp.content)

        if resp.status_code >= 400:
            raise GrocyError(resp)
        if len(resp.content) > 0:
            return resp.json()
        return None

    def _do_put_request(self, end_url: str, data):
        req_url = urljoin(self._base_url, end_url)
        up_header = self._headers.copy()
        up_header["accept"] = "*/*"
        if isinstance(data, dict):
            up_header["Content-Type"] = "application/json"
            data = json.dumps(data)
        else:
            up_header["Content-Type"] = "application/octet-stream"
        resp = requests.put(
            req_url, verify=self._verify_ssl, headers=up_header, data=data
        )

        _LOGGER.debug("-->\tPUT /%s", end_url)
        _LOGGER.debug("\t\t%s", data)
        _LOGGER.debug("<--\t%d for /%s", resp.status_code, end_url)
        _LOGGER.debug("\t\t%s", resp.content)

        if resp.status_code >= 400:
            raise GrocyError(resp)

        if len(resp.content) > 0:
            return resp.json()
        return None

    def _do_delete_request(self, end_url: str):
        req_url = urljoin(self._base_url, end_url)
        resp = requests.delete(req_url, verify=self._verify_ssl, headers=self._headers)

        _LOGGER.debug("-->\tDELETE /%s", end_url)
        _LOGGER.debug("<--\t%d for /%s", resp.status_code, end_url)
        _LOGGER.debug("\t\t%s", resp.content)

        if resp.status_code >= 400:
            raise GrocyError(resp)

        if len(resp.content) > 0:
            return resp.json()
        return None

    def get_stock(self) -> list[CurrentStockResponse]:
        parsed_json = self._do_get_request("stock")
        if parsed_json:
            return [CurrentStockResponse(**response) for response in parsed_json]
        return []

    def get_volatile_stock(self) -> CurrentVolatilStockResponse:
        parsed_json = self._do_get_request("stock/volatile")
        return CurrentVolatilStockResponse(**parsed_json)

    def get_product(self, product_id) -> ProductDetailsResponse:
        url = f"stock/products/{product_id}"
        parsed_json = self._do_get_request(url)
        if parsed_json:
            return ProductDetailsResponse(**parsed_json)
        return None

    def get_product_by_barcode(self, barcode) -> ProductDetailsResponse:
        url = f"stock/products/by-barcode/{barcode}"
        parsed_json = self._do_get_request(url)
        if parsed_json:
            return ProductDetailsResponse(**parsed_json)
        return None

    def get_chores(
        self, query_filters: list[str] | None = None
    ) -> list[CurrentChoreResponse]:
        parsed_json = self._do_get_request("chores", query_filters)
        if parsed_json:
            return [CurrentChoreResponse(**chore) for chore in parsed_json]
        return []

    def get_chore(self, chore_id: int) -> ChoreDetailsResponse:
        url = f"chores/{chore_id}"
        parsed_json = self._do_get_request(url)
        if parsed_json:
            return ChoreDetailsResponse(**parsed_json)
        return None

    def execute_chore(
        self,
        chore_id: int,
        done_by: int | None = None,
        tracked_time: datetime | None = None,
        skipped: bool = False,
    ):
        if tracked_time is None:
            tracked_time = datetime.now()

        localized_tracked_time = localize_datetime(tracked_time)

        data = {
            "tracked_time": grocy_datetime_str(localized_tracked_time),
            "skipped": skipped,
        }

        if done_by is not None:
            data["done_by"] = done_by

        return self._do_post_request(f"chores/{chore_id}/execute", data)

    def add_product(
        self,
        product_id,
        amount: float,
        price: float,
        best_before_date: datetime | None = None,
        transaction_type: TransactionType = TransactionType.PURCHASE,
    ):
        data = {
            "amount": amount,
            "transaction_type": transaction_type.value,
            "price": price,
        }

        if best_before_date is not None:
            data["best_before_date"] = best_before_date.strftime("%Y-%m-%d")

        return self._do_post_request(f"stock/products/{product_id}/add", data)

    def consume_product(
        self,
        product_id: int,
        amount: float = 1,
        spoiled: bool = False,
        transaction_type: TransactionType = TransactionType.CONSUME,
        allow_subproduct_substitution: bool = False,
    ):
        data = {
            "amount": amount,
            "spoiled": spoiled,
            "transaction_type": transaction_type.value,
            "allow_subproduct_substitution": allow_subproduct_substitution,
        }

        self._do_post_request(f"stock/products/{product_id}/consume", data)

    def open_product(
        self,
        product_id: int,
        amount: float = 1,
        allow_subproduct_substitution: bool = False,
    ):
        data = {
            "amount": amount,
            "allow_subproduct_substitution": allow_subproduct_substitution,
        }

        self._do_post_request(f"stock/products/{product_id}/open", data)

    def consume_recipe(
        self,
        recipe_id: int,
    ):
        self._do_post_request(f"recipes/{recipe_id}/consume", None)

    def inventory_product(
        self,
        product_id: int,
        new_amount: float,
        best_before_date: datetime = None,
        shopping_location_id: int = None,
        location_id: int = None,
        price: float = None,
    ):
        data = {
            "new_amount": new_amount,
        }

        if best_before_date is not None:
            data["best_before_date"] = localize_datetime(best_before_date).strftime(
                "%Y-%m-%d"
            )
        if shopping_location_id is not None:
            data["shopping_location_id"] = shopping_location_id

        if location_id is not None:
            data["location_id"] = location_id

        if price is not None:
            data["price"] = price

        parsed_json = self._do_post_request(
            f"stock/products/{product_id}/inventory", data
        )

        if parsed_json:
            stockLog = [StockLogResponse(**response) for response in parsed_json]
            return stockLog[0]
        return None

    def add_product_by_barcode(
        self,
        barcode: str,
        amount: float,
        price: float,
        best_before_date: datetime | None = None,
    ) -> StockLogResponse:
        data = {
            "amount": amount,
            "transaction_type": TransactionType.PURCHASE.value,
            "price": price,
        }

        if best_before_date is not None:
            data["best_before_date"] = localize_datetime(best_before_date).strftime(
                "%Y-%m-%d"
            )

        parsed_json = self._do_post_request(
            f"stock/products/by-barcode/{barcode}/add", data
        )

        if parsed_json:
            stockLog = [StockLogResponse(**response) for response in parsed_json]
            return stockLog[0]
        return None

    def consume_product_by_barcode(
        self, barcode: str, amount: float = 1, spoiled: bool = False
    ):
        data = {
            "amount": amount,
            "spoiled": spoiled,
            "transaction_type": TransactionType.CONSUME.value,
        }

        parsed_json = self._do_post_request(
            f"stock/products/by-barcode/{barcode}/consume", data
        )

        if parsed_json:
            stockLog = [StockLogResponse(**response) for response in parsed_json]
            return stockLog[0]
        return None

    def inventory_product_by_barcode(
        self,
        barcode: str,
        new_amount: float,
        best_before_date: datetime | None = None,
        location_id: int | None = None,
        price: float | None = None,
    ):
        data = {
            "new_amount": new_amount,
        }

        if best_before_date is not None:
            data["best_before_date"] = localize_datetime(best_before_date).strftime(
                "%Y-%m-%d"
            )

        if location_id is not None:
            data["location_id"] = location_id

        if price is not None:
            data["price"] = price

        parsed_json = self._do_post_request(
            f"stock/products/by-barcode/{barcode}/inventory", data
        )

        if parsed_json:
            stockLog = [StockLogResponse(**response) for response in parsed_json]
            return stockLog[0]
        return None

    def get_shopping_list(
        self, query_filters: list[str] = None
    ) -> list[ShoppingListItem]:
        parsed_json = self._do_get_request("objects/shopping_list", query_filters)
        if parsed_json:
            return [ShoppingListItem(**response) for response in parsed_json]
        return []

    def add_missing_product_to_shopping_list(self, shopping_list_id: int = None):
        data = None
        if shopping_list_id:
            data = {"list_id": shopping_list_id}

        self._do_post_request("stock/shoppinglist/add-missing-products", data)

    def add_product_to_shopping_list(
        self,
        product_id: int,
        shopping_list_id: int = 1,
        amount: float = 1,
        quantity_unit_id: int = None,
    ):
        data = {
            "product_id": product_id,
            "list_id": shopping_list_id,
            "product_amount": amount,
        }
        if quantity_unit_id:
            data["qu_id"] = quantity_unit_id
        self._do_post_request("stock/shoppinglist/add-product", data)

    def clear_shopping_list(self, shopping_list_id: int = 1):
        data = {"list_id": shopping_list_id}

        self._do_post_request("stock/shoppinglist/clear", data)

    def remove_product_in_shopping_list(
        self, product_id: int, shopping_list_id: int = 1, amount: float = 1
    ):
        data = {
            "product_id": product_id,
            "list_id": shopping_list_id,
            "product_amount": amount,
        }
        self._do_post_request("stock/shoppinglist/remove-product", data)

    def get_product_groups(
        self, query_filters: list[str] | None = None
    ) -> list[LocationData]:
        parsed_json = self._do_get_request("objects/product_groups", query_filters)
        if parsed_json:
            return [LocationData(**response) for response in parsed_json]
        return []

    def upload_product_picture(self, product_id: int, pic_path: str):
        b64fn = base64.b64encode(f"{product_id}.jpg".encode("ascii"))
        req_url = "files/productpictures/" + str(b64fn, "utf-8")
        with open(pic_path, "rb") as pic:  # noqa: PTH123
            self._do_put_request(req_url, pic)

    def update_product_pic(self, product_id: int):
        pic_name = f"{product_id}.jpg"
        data = {"picture_file_name": pic_name}
        self._do_put_request(f"objects/products/{product_id}", data)

    def get_userfields(self, entity: str, object_id: int):
        url = f"userfields/{entity}/{object_id}"
        return self._do_get_request(url)

    def set_userfields(self, entity: str, object_id: int, key: str, value):
        data = {key: value}
        self._do_put_request(f"userfields/{entity}/{object_id}", data)

    def get_last_db_changed(self):
        resp = self._do_get_request("system/db-changed-time")
        return parse_date(resp.get("changed_time"))

    def get_system_info(self) -> SystemInfoDto:
        parsed_json = self._do_get_request("system/info")
        if parsed_json:
            return SystemInfoDto(**parsed_json)

    def get_system_time(self) -> SystemTimeDto:
        parsed_json = self._do_get_request("system/time")
        if parsed_json:
            return SystemTimeDto(**parsed_json)

    def get_system_config(self) -> SystemConfigDto:
        parsed_json = self._do_get_request("system/config")
        _LOGGER.debug("System config: %s", parsed_json)
        if parsed_json:
            return SystemConfigDto(**parsed_json)

    def get_tasks(self, query_filters: list[str] | None = None) -> list[TaskResponse]:
        parsed_json = self._do_get_request("tasks", query_filters)
        if parsed_json:
            return [TaskResponse(**data) for data in parsed_json]
        return []

    def get_task(self, task_id: int) -> TaskResponse:
        url = f"objects/tasks/{task_id}"
        parsed_json = self._do_get_request(url)
        return TaskResponse(**parsed_json)

    def complete_task(self, task_id: int, done_time: datetime | None = None):
        url = f"tasks/{task_id}/complete"

        if done_time is None:
            done_time = datetime.now()

        localized_done_time = localize_datetime(done_time)

        data = {"done_time": grocy_datetime_str(localized_done_time)}
        self._do_post_request(url, data)

    def get_meal_plan(
        self, query_filters: list[str] | None = None
    ) -> list[MealPlanResponse]:
        parsed_json = self._do_get_request("objects/meal_plan", query_filters)
        if parsed_json:
            return [MealPlanResponse(**data) for data in parsed_json]
        return []

    def get_recipe(self, object_id: int) -> RecipeDetailsResponse:
        parsed_json = self._do_get_request(f"objects/recipes/{object_id}")
        if parsed_json:
            return RecipeDetailsResponse(**parsed_json)
        return None

    def get_batteries(
        self, query_filters: list[str] | None = None
    ) -> list[CurrentBatteryResponse]:
        parsed_json = self._do_get_request("batteries", query_filters)
        if parsed_json:
            return [CurrentBatteryResponse(**data) for data in parsed_json]
        return []

    def get_battery(self, battery_id: int) -> BatteryDetailsResponse:
        parsed_json = self._do_get_request(f"batteries/{battery_id}")
        if parsed_json:
            return BatteryDetailsResponse(**parsed_json)
        return None

    def charge_battery(self, battery_id: int, tracked_time: datetime | None = None):
        if tracked_time is None:
            tracked_time = datetime.now()

        localized_tracked_time = localize_datetime(tracked_time)
        data = {"tracked_time": grocy_datetime_str(localized_tracked_time)}

        return self._do_post_request(f"batteries/{battery_id}/charge", data)

    def add_generic(self, entity_type: str, data):
        return self._do_post_request(f"objects/{entity_type}", data)

    def get_generic(self, entity_type: str, object_id: int):
        return self._do_get_request(f"objects/{entity_type}/{object_id}")

    def update_generic(self, entity_type: str, object_id: int, data):
        return self._do_put_request(f"objects/{entity_type}/{object_id}", data)

    def delete_generic(self, entity_type: str, object_id: int):
        return self._do_delete_request(f"objects/{entity_type}/{object_id}")

    def get_generic_objects_for_type(
        self, entity_type: str | Enum, query_filters: list[str] | None = None
    ):
        if isinstance(entity_type, Enum):
            entity_type = entity_type.value
        return self._do_get_request(f"objects/{entity_type}", query_filters)

    def get_meal_plan_sections(
        self, query_filters: list[str] | None = None
    ) -> list[MealPlanSectionResponse]:
        parsed_json = self.get_generic_objects_for_type(
            EntityType.MEAL_PLAN_SECTIONS, query_filters
        )
        if parsed_json:
            return [MealPlanSectionResponse(**resp) for resp in parsed_json]
        return []

    def get_meal_plan_section(self, meal_plan_section_id) -> MealPlanSectionResponse:
        parsed_json = self._do_get_request(
            f"objects/meal_plan_sections?query%5B%5D=id%3D{meal_plan_section_id}"
        )
        if parsed_json and len(parsed_json) == 1:
            return MealPlanSectionResponse(**parsed_json[0])
        return None

    def get_equipment(self, equipment_id: int) -> EquipmentDetailsResponse:
        """Get details of a specific equipment item."""
        parsed_json = self._do_get_request(f"objects/equipment/{equipment_id}")
        if parsed_json:
            equipment_data = EquipmentData(**parsed_json)
            return EquipmentDetailsResponse(equipment=equipment_data)
        return None

    def get_all_equipment(self, query_filters: list[str] = None) -> list[dict]:
        """Get all equipment items."""
        return self._do_get_request("objects/equipment", query_filters) or []

    def get_users(self) -> list[UserDto]:
        parsed_json = self._do_get_request("users")
        if parsed_json:
            return [UserDto(**user) for user in parsed_json]
        return []

    def get_user(self, user_id: int) -> UserDto:
        query_params = []
        if user_id:
            query_params.append(f"id={user_id}")
        parsed_json = self._do_get_request("users")
        if parsed_json:
            return UserDto(**parsed_json[0])
        return None



from datetime import datetime

from pygrocy2.base import DataModel
from pygrocy2.grocy_api_client import (
    EquipmentDetailsResponse,
    EquipmentData,
    GrocyApiClient,
)


class Equipment(DataModel):
    """Class to represent a Grocy equipment item."""

    def __init__(self, response):
        self._init_empty()

        if isinstance(response, EquipmentData):
            self._init_from_EquipmentData(response)
        elif isinstance(response, EquipmentDetailsResponse):
            self._init_from_EquipmentDetailsResponse(response)
        elif isinstance(response, dict):
            # Initialize from a plain dictionary response from the API
            self._init_from_dict(response)

    def _init_from_EquipmentData(self, response: EquipmentData):
        self._id = response.id
        self._name = response.name
        self._description = response.description
        self._instruction_manual_file_name = response.instruction_manual_file_name
        self._created_timestamp = response.created_timestamp
        self._userfields = response.userfields

    def _init_from_EquipmentDetailsResponse(self, response: EquipmentDetailsResponse):
        self._id = response.equipment.id
        self._name = response.equipment.name
        self._description = response.equipment.description
        self._instruction_manual_file_name = (
            response.equipment.instruction_manual_file_name
        )
        self._created_timestamp = response.equipment.created_timestamp
        self._userfields = response.equipment.userfields

    def _init_from_dict(self, response: dict):
        self._id = response.get("id")
        self._name = response.get("name")
        self._description = response.get("description")
        self._instruction_manual_file_name = response.get(
            "instruction_manual_file_name"
        )
        self._created_timestamp = response.get("row_created_timestamp") or response.get(
            "created_timestamp"
        )
        self._userfields = response.get("userfields")

    def _init_empty(self):
        self._id = None
        self._name = None
        self._description = None
        self._instruction_manual_file_name = None
        self._created_timestamp = None
        self._userfields = None

    def get_details(self, api_client: GrocyApiClient):
        details = api_client.get_equipment(self._id)
        self._init_from_EquipmentDetailsResponse(details)

    @property
    def id(self) -> int:
        return self._id

    @property
    def name(self) -> str:
        return self._name

    @property
    def description(self) -> str:
        return self._description

    @property
    def instruction_manual_file_name(self) -> str:
        return self._instruction_manual_file_name

    @property
    def created_timestamp(self) -> datetime:
        return self._created_timestamp

    @property
    def userfields(self):
        return self._userfields

    def as_dict(self):
        """Return a dict representation of the equipment."""
        return {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "instruction_manual_file_name": self.instruction_manual_file_name,
            "created_timestamp": self.created_timestamp,
            "userfields": self.userfields,
        }



import datetime

import pytest

from pygrocy2.errors import GrocyError


class TestMealPlanSections:
    @pytest.mark.vcr
    def test_get_sections_valid(self, grocy):
        sections = grocy.meal_plan_sections()

        assert len(sections) == 4
        section = sections[1]
        assert section.id == 1
        assert section.name == "Breakfast"
        assert section.sort_number == 10
        assert isinstance(section.row_created_timestamp, datetime.datetime)

    @pytest.mark.vcr
    def test_get_section_by_id_valid(self, grocy):
        section = grocy.meal_plan_section(1)

        assert section.id == 1

    @pytest.mark.vcr
    def test_get_section_by_id_invalid(self, grocy):
        section = grocy.meal_plan_section(1000)
        assert section is None

    @pytest.mark.vcr
    def test_get_sections_filters_valid(self, grocy):
        query_filter = ["name=Breakfast"]
        sections = grocy.meal_plan_sections(query_filters=query_filter)

        for item in sections:
            assert item.name == "Breakfast"

    @pytest.mark.vcr
    def test_get_sections_filters_invalid(self, grocy, invalid_query_filter):
        with pytest.raises(GrocyError) as exc_info:
            grocy.meal_plan_sections(query_filters=invalid_query_filter)

        error = exc_info.value
        assert error.status_code == 500



from datetime import datetime

import pytest

from pygrocy2.data_models.equipment import Equipment
from pygrocy2.errors import GrocyError
from pygrocy2.grocy import Grocy
from pygrocy2.grocy_api_client import (
    EquipmentDetailsResponse,
    EquipmentData,
)


class TestEquipment:
    def test_init_from_equipment_data(self):
        """Test Equipment initialization from EquipmentData with minimal fields."""
        response = EquipmentData(
            id=1, name="Test Equipment", row_created_timestamp=datetime(2025, 3, 23)
        )
        equipment = Equipment(response)

        assert equipment.id == 1
        assert equipment.name == "Test Equipment"
        assert equipment.description is None
        assert equipment.instruction_manual_file_name is None
        assert equipment.created_timestamp == datetime(2025, 3, 23)
        assert equipment.userfields is None

    def test_init_from_details_response(self):
        """Test Equipment initialization from EquipmentDetailsResponse."""
        equipment_data = EquipmentData(
            id=1,
            name="Test Equipment",
            description="Test Description",
            instruction_manual_file_name="manual.pdf",
            row_created_timestamp=datetime(2025, 3, 23),
            userfields={"custom_field": "custom_value"},
        )

        response = EquipmentDetailsResponse(equipment=equipment_data)

        equipment = Equipment(response)

        assert equipment.id == 1
        assert equipment.name == "Test Equipment"
        assert equipment.description == "Test Description"
        assert equipment.instruction_manual_file_name == "manual.pdf"
        assert equipment.created_timestamp == datetime(2025, 3, 23)
        assert equipment.userfields == {"custom_field": "custom_value"}

    def test_as_dict(self):
        """Test Equipment.as_dict method."""
        equipment_data = EquipmentData(
            id=1,
            name="Test Equipment",
            description="Test Description",
            instruction_manual_file_name="manual.pdf",
            row_created_timestamp=datetime(2025, 3, 23),
            userfields={"custom_field": "custom_value"},
        )

        response = EquipmentDetailsResponse(equipment=equipment_data)

        equipment = Equipment(response)
        result = equipment.as_dict()

        expected = {
            "id": 1,
            "name": "Test Equipment",
            "description": "Test Description",
            "instruction_manual_file_name": "manual.pdf",
            "created_timestamp": datetime(2025, 3, 23),
            "userfields": {"custom_field": "custom_value"},
        }
        assert result == expected

    @pytest.mark.vcr
    def test_get_equipment_list(self, grocy: Grocy):
        equipment = grocy.equipment(get_details=False)

        assert isinstance(equipment, list)
        assert len(equipment) == 2
        for item in equipment:
            assert isinstance(item, Equipment)
            assert isinstance(item.id, int)
            assert isinstance(item.name, str)

        item = next(e for e in equipment if e.id == 1)
        assert item.name == "Coffee machine"
        assert item.instruction_manual_file_name == "loremipsum.pdf"

    @pytest.mark.vcr
    def test_get_equipment_with_details(self, grocy: Grocy):
        equipment = grocy.equipment(get_details=True)

        assert len(equipment) == 2

        item = next(e for e in equipment if e.id == 1)
        assert item.name == "Coffee machine"
        assert item.description is not None
        assert item.instruction_manual_file_name == "loremipsum.pdf"
        assert isinstance(item.created_timestamp, datetime)
        assert item.userfields is None

        dishwasher = next(e for e in equipment if e.id == 2)
        assert dishwasher.name == "Dishwasher"
        assert dishwasher.instruction_manual_file_name is None

    @pytest.mark.vcr
    def test_get_equipment_details(self, grocy: Grocy):
        equipment = grocy.get_equipment(1)

        assert isinstance(equipment, Equipment)
        assert equipment.id == 1
        assert equipment.name == "Coffee machine"
        assert equipment.description is not None
        assert equipment.description.startswith("<h1>Lorem ipsum")
        assert equipment.instruction_manual_file_name == "loremipsum.pdf"
        assert isinstance(equipment.created_timestamp, datetime)
        assert equipment.userfields is None

    @pytest.mark.vcr
    def test_get_equipment_details_non_existent(self, grocy: Grocy):
        with pytest.raises(GrocyError) as exc_info:
            grocy.get_equipment(999)

        error = exc_info.value
        assert error.status_code == 404

    @pytest.mark.vcr
    def test_get_equipment_filters_valid(self, grocy: Grocy):
        query_filter = ["name=Coffee machine"]
        equipment = grocy.equipment(query_filters=query_filter)

        assert len(equipment) == 1
        assert equipment[0].name == "Coffee machine"

    @pytest.mark.vcr
    def test_get_equipment_filters_invalid(self, grocy: Grocy, invalid_query_filter):
        with pytest.raises(GrocyError) as exc_info:
            grocy.equipment(query_filters=invalid_query_filter)

        error = exc_info.value
        assert error.status_code == 500



from datetime import datetime

import pytest

from pygrocy2.data_models.chore import AssignmentType, Chore, PeriodType
from pygrocy2.data_models.user import User
from pygrocy2.errors import GrocyError


class TestChores:
    @pytest.mark.vcr
    def test_get_chores_valid(self, grocy):
        chores = grocy.chores(get_details=True)

        assert isinstance(chores, list)
        assert len(chores) == 6
        for chore in chores:
            assert isinstance(chore, Chore)
            assert isinstance(chore.id, int)
            assert isinstance(chore.last_tracked_time, datetime)
            assert isinstance(chore.next_estimated_execution_time, datetime)
            assert isinstance(chore.name, str)
            assert isinstance(chore.last_done_by, User)
            assert chore.userfields is not None
            assert isinstance(chore.userfields, (list, dict))

        chore = next(chore for chore in chores if chore.id == 6)
        assert chore.name == "Change the bed sheets"
        assert chore.period_config == "monday"

    @pytest.mark.vcr
    def test_get_chore_details(self, grocy):
        chore_details = grocy.chore(3)
        assert isinstance(chore_details, Chore)

        assert chore_details.name == "Take out the trash"
        assert chore_details.assignment_type == AssignmentType.RANDOM
        assert chore_details.last_done_by.id == 3
        assert chore_details.period_type == PeriodType.HOURLY
        assert chore_details.period_days is None
        assert chore_details.period_config is None
        assert chore_details.track_date_only is True
        assert chore_details.rollover is False
        assert chore_details.assignment_config == "1,2,3,4"
        assert chore_details.next_execution_assigned_user.id == 2
        assert chore_details.next_execution_assigned_to_user_id == 2
        assert chore_details.userfields is None

    @pytest.mark.vcr
    def test_execute_chore_valid(self, grocy):
        result = grocy.execute_chore(1)
        assert not isinstance(result, GrocyError)

    @pytest.mark.vcr
    def test_execute_chore_valid_with_data(self, grocy):
        result = grocy.execute_chore(
            1, done_by=1, tracked_time=datetime.now(), skipped=False
        )
        assert not isinstance(result, GrocyError)

    @pytest.mark.vcr
    def test_execute_chore_invalid(self, grocy):
        with pytest.raises(GrocyError) as exc_info:
            grocy.execute_chore(1000)

        error = exc_info.value
        assert error.status_code == 400

    @pytest.mark.vcr
    def test_get_chores_filters_valid(self, grocy):
        query_filter = ["next_execution_assigned_to_user_id=1"]
        chores = grocy.chores(get_details=True, query_filters=query_filter)

        for item in chores:
            assert item.next_execution_assigned_to_user_id == 1

    @pytest.mark.vcr
    def test_get_chores_filters_invalid(self, grocy, invalid_query_filter):
        with pytest.raises(GrocyError) as exc_info:
            grocy.chores(get_details=True, query_filters=invalid_query_filter)

        error = exc_info.value
        assert error.status_code == 500



from enum import Enum


class EntityType(str, Enum):
    PRODUCTS = "products"
    CHORES = "chores"
    PRODUCT_BARCODES = "product_barcodes"
    BATTERIES = "batteries"
    LOCATIONS = "locations"
    QUANTITY_UNITS = "quantity_units"
    QUANTITY_UNIT_CONVERSIONS = "quantity_unit_conversions"
    SHOPPING_LIST = "shopping_list"
    SHOPPING_LISTS = "shopping_lists"
    SHOPPING_LOCATIONS = "shopping_locations"
    RECIPES = "recipes"
    RECIPES_POS = "recipes_pos"
    RECIPES_NESTINGS = "recipes_nestings"
    TASKS = "tasks"
    TASK_CATEGORIES = "task_categories"
    PRODUCT_GROUPS = "product_groups"
    EQUIPMENT = "equipment"
    USER_FIELDS = "userfields"
    USER_ENTITIES = "userentities"
    USER_OBJECTS = "userobjects"
    MEAL_PLAN = "meal_plan"
    MEAL_PLAN_SECTIONS = "meal_plan_sections"



from pygrocy2.data_models.product import ProductBarcode
from pygrocy2.grocy_api_client import ProductBarcodeData


class TestMisc:
    def test_158_productbarcode_deserialization(self):
        parsed_data = {"barcode": "123"}
        data = ProductBarcodeData(**parsed_data)

        barcode = ProductBarcode(data)
        result = barcode.toJson()

        assert result is not None
        assert "barcode" in result
        assert "123" in result



"""Legacy setuptools fallback.

The project uses pyproject.toml for metadata and builds.
"""

from setuptools import setup


if __name__ == "__main__":  # pragma: no cover - compatibility shim only
    setup()



from datetime import datetime
import zoneinfo

from tzlocal import get_localzone


def parse_date(input_value):
    if input_value == "" or input_value is None:
        return None
    return datetime.fromisoformat(input_value)


def parse_int(input_value, default_value=None):
    if input_value is None:
        return default_value
    try:
        return int(input_value)
    except ValueError:
        return default_value


def parse_float(input_value, default_value=None):
    if input_value is None:
        return default_value
    try:
        return float(input_value)
    except ValueError:
        return default_value


def parse_bool_int(input_value):
    if input_value is None:
        return False
    try:
        num = int(input_value)
        return bool(num)
    except ValueError:
        return False


def localize_datetime(timestamp: datetime) -> datetime:
    if timestamp.tzinfo is not None:
        return timestamp

    local_zone = _detect_local_zone()
    if local_zone is not None:
        # Treat naive input as already in local time so we just attach tzinfo.
        return timestamp.replace(tzinfo=local_zone)

    return timestamp.astimezone()


def grocy_datetime_str(timestamp: datetime) -> str:
    if timestamp is None:
        return ""
    return timestamp.strftime("%Y-%m-%d %H:%M:%S")


def _detect_local_zone():
    local_zone = get_localzone()
    if isinstance(local_zone, str):
        return zoneinfo.ZoneInfo(local_zone)
    return local_zone



from datetime import date, datetime

from pygrocy2.base import DataModel
from pygrocy2.grocy_api_client import SystemConfigDto, SystemInfoDto, SystemTimeDto


class SystemInfo(DataModel):
    def __init__(self, system_info_dto: SystemInfoDto):
        self._grocy_version = system_info_dto.grocy_version_info.version
        self._grocy_release_date = system_info_dto.grocy_version_info.release_date
        self._php_version = system_info_dto.php_version
        self._sqlite_version = system_info_dto.sqlite_version
        self._os = system_info_dto.os
        self._client = system_info_dto.client

    @property
    def grocy_version(self) -> str:
        return self._grocy_version

    @property
    def grocy_release_date(self) -> date:
        return self._grocy_release_date

    @property
    def php_version(self) -> str:
        return self._php_version

    @property
    def sqlite_version(self) -> str:
        return self._sqlite_version

    @property
    def os(self) -> str:
        return self._os

    @property
    def client(self) -> str:
        return self._client


class SystemTime(DataModel):
    def __init__(self, system_time_dto: SystemTimeDto):
        self._timezone = system_time_dto.timezone
        self._time_local = system_time_dto.time_local
        self._time_local_sqlite3 = system_time_dto.time_local_sqlite3
        self._time_utc = system_time_dto.time_utc
        self._timestamp = system_time_dto.timestamp

    @property
    def timezone(self) -> str:
        return self._timezone

    @property
    def time_local(self) -> datetime:
        return self._time_local

    @property
    def time_local_sqlite3(self) -> datetime:
        return self._time_local_sqlite3

    @property
    def time_utc(self) -> datetime:
        return self._time_utc

    @property
    def timestamp(self) -> int:
        return self._timestamp


class SystemConfig(DataModel):
    def __init__(self, system_config_dto: SystemConfigDto):
        self._username = system_config_dto.username
        self._base_path = system_config_dto.base_path
        self._base_url = system_config_dto.base_url
        self._mode = system_config_dto.mode
        self._default_locale = system_config_dto.default_locale
        self._locale = system_config_dto.locale
        self._currency = system_config_dto.currency

        self._enabled_features = [
            feature
            for feature, value in system_config_dto.feature_flags.items()
            if value
            not in (False, "0")  # The default is enabled, disabled can be False or "0"
        ]

    @property
    def username(self) -> str:
        return self._username

    @property
    def base_path(self) -> str:
        return self._base_path

    @property
    def base_url(self) -> str:
        return self._base_url

    @property
    def mode(self) -> str:
        return self._mode

    @property
    def default_locale(self) -> str:
        return self._default_locale

    @property
    def locale(self) -> str:
        return self._locale

    @property
    def currency(self) -> str:
        return self._currency

    @property
    def enabled_features(self) -> list[str]:
        return self._enabled_features



import pytest

from pygrocy2.data_models.generic import EntityType
from pygrocy2.errors import GrocyError


class TestGeneric:
    @pytest.mark.vcr
    def test_generic_add_valid(self, grocy):
        data = {"name": "Testbattery"}

        grocy.add_generic(EntityType.BATTERIES, data)

    @pytest.mark.vcr
    def test_generic_add_invalid(self, grocy):
        data = {"eman": "Testbattery"}

        with pytest.raises(GrocyError) as exc_info:
            grocy.add_generic(EntityType.BATTERIES, data)

        error = exc_info.value
        assert error.status_code == 400

    @pytest.mark.vcr
    def test_generic_update_valid(self, grocy):
        updated_data = {"name": "Le new battery"}

        grocy.update_generic(EntityType.BATTERIES, 1, updated_data)

    @pytest.mark.vcr
    def test_generic_update_invalid_id(self, grocy):
        updated_data = {"name": "Le new battery"}

        with pytest.raises(GrocyError) as exc_info:
            grocy.update_generic(EntityType.BATTERIES, 1000, updated_data)

        error = exc_info.value
        assert error.status_code == 400
        assert error.message == "Object not found"

    @pytest.mark.vcr
    def test_generic_update_invalid_data(self, grocy):
        with pytest.raises(GrocyError) as exc_info:
            grocy.update_generic(EntityType.BATTERIES, 1, None)

        error = exc_info.value
        assert error.status_code == 400
        assert error.message == "Bad Content-Type"

    @pytest.mark.vcr
    def test_delete_generic_success(self, grocy):
        grocy.delete_generic(EntityType.TASKS, 3)

    @pytest.mark.vcr
    def test_delete_generic_error(self, grocy):
        with pytest.raises(GrocyError) as exc_info:
            grocy.delete_generic(EntityType.TASKS, 30000)

        error = exc_info.value
        assert error.status_code == 400

    @pytest.mark.vcr
    def test_get_generic_objects_for_type_filters_valid(self, grocy):
        query_filter = ["name=Walmart"]
        shopping_locations = grocy.get_generic_objects_for_type(
            EntityType.SHOPPING_LOCATIONS, query_filters=query_filter
        )

        assert len(shopping_locations) == 0

    @pytest.mark.vcr
    def test_get_generic_objects_for_type_filters_invalid(
        self, grocy, invalid_query_filter
    ):
        with pytest.raises(GrocyError) as exc_info:
            grocy.get_generic_objects_for_type(
                EntityType.SHOPPING_LOCATIONS, query_filters=invalid_query_filter
            )

        error = exc_info.value
        assert error.status_code == 500
